/*
 *  body.c
 *  gmc4asm
 *
 *  Created by Koyama Yoshiaki on 09/08/31.
 *  Copyright 2009 __MyCompanyName__. All rights reserved.
 *
 */

#include "gmc4asm.h"

/*
 
 */

Opcode *Top = NULL;
Symrec *Table = NULL;
int line_number = 0;


char *error_msg[] = {
	"%d 行目: ラベル %s が再定義されています。\n",
	"%d 行目: %s が定義されていません。\n",
	"%d 行目: 60番地以上のアドレスにデータは書き込めません。\n",
	NULL
};

/*
 
 */
void add_top(Opcode *opcode) {
	Opcode *last;

	
	if (Top == NULL) {
		Top = opcode;
	} else {
		last = Top;
		while (last->next != NULL) {
			last = last->next;
		}
		last->next = opcode;
	}
	
}

/*
 
 */
char *name_malloc(char *name) {
	size_t len;
	char *ptr = NULL;
	
	len = strlen(name);

	if ((len != 0) && (len != -1)) {
		ptr = malloc(len+1);
		if (ptr != NULL) {
			strcpy(ptr, name);
			ptr[len] = '\0';
			strlower(ptr);
		}
	}

	return ptr;
}

/*
 
 */
void name_free(char *name) {
	free(name);
}

/*
 
 */
void strlower(char *name) {
	size_t i;
	
	for (i=0; i<strlen(name); i++) {
		name[i] = tolower((int)name[i]);
	}
}

/*

 */
char *name_label_malloc(char *name) {
	char *ptr, *ref;
	size_t len, name_len;
	int i;
	
	len = name_len = strlen(name);
	
	if (name[len-1] == ':') len--;
	if (name[0] == '*') len--;
	
	ref = ptr = malloc(len+1);


	if (ref != NULL) {
		for (i=0; i<name_len; i++) {
			if (name[i] == ':') {
			} else if (name[i] == '*') {
			} else {
				*ptr = name[i];
				ptr++;
			}	
		}
		*ptr = '\0';
		
		strlower(ref);
	}
	
	return ref;
}


/*
 
 */
Opcode *opcode_malloc() {
	Opcode *ptr = NULL;

	ptr = malloc(sizeof(Opcode));

	if (ptr != NULL) {
		ptr->label.flag = FALSE;
	}
	
	return ptr;
}

/*
 
 */
void opcode_free(Opcode *opcode) {
	free(opcode);
}

/*
 
 */
void init_opcode(Opcode *ptr, TypeOpcode type, char *name) {
	ptr->type = type;
	ptr->next = NULL;	
	ptr->line_number = 0;
	
	switch (ptr->type) {
		case TYPE_OPCODE1:
			ptr->nibble_count = 1;
			ptr->u.opcode1.name = name;
			ptr->u.opcode1.code1 = getsym(name);			
			break;
		case TYPE_OPCODE2:
			ptr->nibble_count = 2;
			ptr->u.opcode2.name = name;
			ptr->u.opcode2.code1 = getsym(name);
			break;
		case TYPE_CALL:
			ptr->nibble_count = 2;
			ptr->u.opcode_call.code1 = 0xe;
//			ptr->u.opcode_call.code2 = getsym(name);
			break;
		case TYPE_JUMP:
			ptr->nibble_count = 3;
			ptr->u.opcode_jump.code1 = 0xf;
			break;
		case TYPE_LABEL:
			ptr->nibble_count = 0;
			ptr->u.opcode_label.name = name;
			break;
		case TYPE_NIBBLE_DN:
			ptr->nibble_count = 0;
			ptr->u.opcode_nibble_dn.adress = 0;
			ptr->u.opcode_nibble_dn.next = NULL;			
			break;
		case TYPE_ALIGN:
			ptr->nibble_count = 0;
			ptr->u.opcode_align.adress = 0;
			break;
		default:
			break;
	}
}

/*
 
 */
Boolean calc_label() {
	Opcode *ptr;
	unsigned int adress;
	Boolean flag, ret;

	
	ptr = Top;
	adress = 0;
	ret = TRUE;
	
	while (ptr != NULL) {
		if (ptr->type == TYPE_LABEL) {
			flag = putsym(ptr->u.opcode_label.name, adress);			
			if (flag == FALSE) {
				fprintf(stderr, error_msg[0], ptr->line_number, ptr->u.opcode_label.name);
				ret = FALSE;
			}
		} else {

			if (ptr->label.flag == TRUE) {
				flag = putsym(ptr->label.name, adress);
				if (flag == FALSE) {
					fprintf(stderr, error_msg[0], ptr->line_number, ptr->label.name);
					ret = FALSE;
				}
			}

		}
		
		adress += ptr->nibble_count;		
		ptr = ptr->next;

	}
	return ret;
}

/*
 
 */
Boolean set_label() {		
	Opcode *ptr;
	unsigned int adress;
	Boolean ret = TRUE;
	
	ptr = Top;

	if (ptr == NULL) {
		return FALSE;
	}	
	
	do {
		switch (ptr->type) {
			case TYPE_JUMP:
				adress = getsym(ptr->u.opcode_jump.name);
//				printf("jump %x ", adress);
				ptr->u.opcode_jump.code2 = (adress >> 4) & 0xf; 
				ptr->u.opcode_jump.code3 = adress & 0xf;
//				printf("%x %x \n", ptr->u.opcode_jump.code2, ptr->u.opcode_jump.code3);

				if (adress == 0xffff) {
					fprintf(stderr, error_msg[1], ptr->line_number, ptr->u.opcode_jump.name);
					ret = FALSE;
				}
				break;
			default:
				break;
		}
		ptr = ptr->next;
	} while (ptr != NULL);

	return ret;
}

/*
 
 */
Boolean set_align() {
	Opcode *ptr;
	unsigned int adress = 0x50;
	Boolean ret = TRUE;
	Opcode_nibble *nibble;
	
	ptr = Top;

	if (ptr == NULL) {
		return FALSE;
	}	
	
	do {
		switch (ptr->type) {
			case TYPE_ALIGN:
				adress = ptr->u.opcode_align.adress;
				if (align_over(ptr, adress) == FALSE) {
					return FALSE;
				}				
				break;
			case TYPE_NIBBLE_DN:
				if (align_over(ptr, adress) == FALSE) {
					return FALSE;
				}
				ptr->u.opcode_nibble_dn.adress = adress;
				adress++;
				
				nibble = ptr->u.opcode_nibble_dn.next;
				if (nibble != NULL) {
					if (align_over(ptr, adress) == FALSE) {
						return FALSE;
					}
					adress++;
					while (nibble->next != NULL) {
						if (align_over(ptr, adress) == FALSE) {
							return FALSE;
						}						
						nibble = nibble->next;
						adress++;
					}
				}
				break;
			default:
				break;
		}
		ptr = ptr->next;
	} while (ptr != NULL);
	
	return ret;
}

Boolean align_over(Opcode *ptr, unsigned int adress) {
	Boolean ret = TRUE;
	
	if (adress >= STORAGE_MAX) {
		fprintf(stderr, error_msg[2], ptr->line_number);
		ret = FALSE;
	}

	return ret;
}

/*
 
 */
void print_top() {
	Opcode *ptr;
	int num;
	unsigned int adress;
	Opcode_nibble *nibble;

	ptr = Top;
	adress = 0;
	
	while (ptr != NULL) {

//		printf("%05d %02x: ", ptr->line_number, adress);

		if (ptr->type == TYPE_NIBBLE_DN) {
		} else if (ptr->type == TYPE_ALIGN) {
			print_space(3);
		} else {
			printf("%02x: ", adress);
		}
		
		switch (ptr->type) {
			case TYPE_OPCODE1:
				printf("%01x", ptr->u.opcode1.code1);
				print_space(12-1);
				break;
			case TYPE_OPCODE2:
				printf("%01x  %01x", ptr->u.opcode2.code1, ptr->u.opcode2.code2);
				print_space(12-4);
				break;
			case TYPE_CALL:
				printf("%01x  %01x", ptr->u.opcode_call.code1, ptr->u.opcode_call.code2);
				print_space(12-4);
				break;
			case TYPE_JUMP:
				printf("%01x  %01x  %01x", ptr->u.opcode_jump.code1, ptr->u.opcode_jump.code2 ,ptr->u.opcode_jump.code3);
				print_space(12-7);
				break;
			case TYPE_NIBBLE_DN:
				printf("%02x: ", ptr->u.opcode_nibble_dn.adress);
				print_space(12);
				break;
			default:
				break;
		}
		

		if (ptr->label.flag == TRUE) {
			
			printf("%s", ptr->label.name);
			
			num = 15 - strlen(ptr->label.name);
			
			if (num > 0) {
				print_space(num);
			} else {
				print_space(1);
			}
			
		} else {
			print_space(15);
		}	
		
		switch (ptr->type) {
			case TYPE_OPCODE1:
				printf("%s\n", ptr->u.opcode1.name);
				break;
			case TYPE_OPCODE2:
				printf("%s %x\n", ptr->u.opcode2.name, ptr->u.opcode2.code2);
				break;
			case TYPE_CALL:
				printf("call %s\n", ptr->u.opcode_call.name);
				break;
			case TYPE_JUMP:
				printf("jump %s\n", ptr->u.opcode_jump.name);
				break;
			case TYPE_LABEL:
				printf("%s\n", ptr->u.opcode_label.name);
				break;
			case TYPE_NIBBLE_DN:
				nibble = ptr->u.opcode_nibble_dn.next;

				if (nibble == NULL) {
					printf("dn %1x", ptr->u.opcode_nibble_dn.data);
				} else {
					printf("dn %1x, ", ptr->u.opcode_nibble_dn.data);
				}
				while (nibble != NULL) {					
					printf("%1x", nibble->data);
					nibble = nibble->next;
					if (nibble != NULL) {
						printf(", ");
					} 
				}				
				printf("\n");
				break;
			case TYPE_ALIGN:
				printf(".align %02x\n", ptr->u.opcode_align.adress);
				break;
			default:
				break;
		}
		
		adress += ptr->nibble_count;
		ptr = ptr->next;
	}

	printf("\n");
}

/*
 
 */
void print_space(int num) {
	int i;
	
	for (i=0; i<num; i++)
		printf(" ");
}

/*
 
 */
struct Sym_table_tag
{
	char *name;
	unsigned int code;
};

typedef struct Sym_table_tag Sym_table;

/*
	{命令記号、命令コード(16進数)}
*/
Sym_table sym[] =
{
	{ "ka", 0 },
	{ "ao", 1 },
	{ "ch", 2 },
	{ "cy", 3 },
	{ "am", 4 },
	{ "ma", 5 },
	{ "m+", 6 },
	{ "m-", 7 },

	{ "tia", 8 },
	{ "aia", 9 },
	{ "tiy", 0xa },
	{ "aiy", 0xb },
	{ "cia", 0xc },
	{ "ciy", 0xd },
	{ "jump", 0xf },

	{ "rsto", 0 },
	{ "setr", 1 },
	{ "rstr", 2 },
	{ "cmpl", 4 },
	{ "chng", 5 },

	{ "sift", 6 },
	{ "shift", 6 },

	{ "ends", 7 },
	{ "errs", 8 },
	{ "shts", 9 },
	{ "lons", 0xa },
	{ "sund", 0xb },
	{ "timr", 0xc },
	{ "dspr", 0xd },
	{ "dem-", 0xe },
	{ "dem+", 0xf},
	{ 0,0 }
};

/*
 
 */
void init_table() {
	int i;
	Boolean flag;

	for (i = 0; sym[i].name != 0; i++) {		
		flag = putsym(sym[i].name, sym[i].code);
		if (flag == FALSE) {
			fprintf(stderr, "symbol table error\n");
		} else {
//			printf("%02x %s ", sym[i].code, sym[i].name);
		}
	}
}

/*
 
 */
Boolean putsym(char *name, unsigned int code) {
	Symrec *ptr, *sym;

	ptr = Table;

	if (ptr != NULL) {
	
		if (strcmp(ptr->name, name) == 0)
			return FALSE;

		while (ptr->next != NULL) {
			ptr = ptr->next;
			if (strcmp(ptr->name, name) == 0)
				return FALSE;
		}

		sym = malloc(sizeof(Symrec));
		sym->name = name_malloc(name);
		sym->code = code;
		sym->next = NULL;
		
		ptr->next = sym;
		
	} else {
		sym = malloc(sizeof(Symrec));
		sym->name = name_malloc(name);
		sym->code = code;
		sym->next = NULL;
		
		Table = sym;
	}
	
	return TRUE;
}

/*

 */
unsigned int getsym(char *name) {
	Symrec *ptr;

	ptr = Table;

	if (ptr != NULL) {
		do {
			if (strcmp(ptr->name, name) == 0) {
				return ptr->code;
			}
			ptr = ptr->next;
		} while (ptr != NULL);
	}
	return 0xffff;
}

/*

*/
void print_table() {
	Symrec *ptr;
	
	ptr = Table;

	if (ptr == NULL) {
		return;
	} else {
		printf("%02x : %s\n", ptr->code, ptr->name);
	}
		
	do {
		ptr = ptr->next;
		printf("%02x : %s\n", ptr->code, ptr->name);
		
	} while (ptr->next != NULL);
}

/*
 
 */
Boolean set_table(char *name, unsigned int code) {
	Symrec *ptr;
	
	ptr = Table;
	
	do {
		if (strcmp(ptr->name, name) == 0) {
			ptr->code = code;
			return TRUE;
		}	
		ptr = ptr->next;
	} while (ptr != NULL);

	return FALSE;
}


/*
 
 
 */
unsigned int calc_dump(unsigned int *dump) {
	unsigned int i, align;
	unsigned int adress, count;
	Opcode *ptr;
	Opcode_nibble *nibble;
	
	// 
	for (i=0; i<STORAGE_MAX; i++) {
		dump[i] = 0xf;
	}

	
	ptr = Top;
	adress = 0;
	count = 0;

	while (ptr != NULL) {
		switch (ptr->type) {
			case TYPE_OPCODE1:
				dump[count] = ptr->u.opcode1.code1;
				count++;
				break;
			case TYPE_OPCODE2:
				dump[count] = ptr->u.opcode2.code1;
				count++;
				dump[count] = ptr->u.opcode2.code2;
				count++;
//				print_space(12-6);
				break;
			case TYPE_CALL:
				dump[count] = ptr->u.opcode_call.code1;
				count++;
				dump[count] = ptr->u.opcode_call.code2;
				count++;
//				print_space(12-6);
				break;
			case TYPE_JUMP:
				dump[count] = ptr->u.opcode_jump.code1;
				count++;
				dump[count] = ptr->u.opcode_jump.code2;
				count++;
				dump[count] = ptr->u.opcode_jump.code3;
				count++;
//				print_space(12-10);
				break;
			case TYPE_NIBBLE_DN:
				align = ptr->u.opcode_nibble_dn.adress;				
				dump[align] = ptr->u.opcode_nibble_dn.data;
				align++;
				
				nibble = ptr->u.opcode_nibble_dn.next;
				
				while (nibble != NULL) {
					dump[align] = nibble->data; 
					nibble = nibble->next;					
					align++;
				}
//				print_space(12-10);
				break;
			default:
				break;
		}
		ptr = ptr->next;
	}


	return count;
}

/*
 
 */
void print_dump(unsigned int *dump, unsigned int count) {
	unsigned int i;
	
	for (i=0; i<count; i++) {
		if (i == 0) {
			printf("%02x:", i);		
		} else if ((i % 16) == 0) {	
			printf("\n%02x:", i);
		} else if ((i % 8) == 0) {
			printf(" ");		
		}
		printf(" %01x", dump[i]);
	}	
	printf("\n\n");
}

/*
 
 */
void print_fxp(FILE *fp, unsigned int *dump, unsigned int count) {
	unsigned int i;
	

	fprintf(fp, "<?xml version=\"1.0\"?>\r\n");
	fprintf(fp, "<ArrayOfString xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\">\r\n");

	for (i=0; i<STORAGE_MAX; i++) {
		fprintf(fp, "  <string>%01x</string>\r\n", dump[i]);
	}	
	fprintf(fp, "</ArrayOfString>\r\n");
}

/*
 
 */
void print_hex(FILE *fp, unsigned int *dump, unsigned int count) {
	unsigned int i;
	
	for (i=0; i<STORAGE_MAX; i++) {
		if (i == 0) {
			fprintf(fp, "%1X", dump[i]);
		} else if ((i % 16) == 0) {
			fprintf(fp, "\r\n%1X", dump[i]);
		} else {
			fprintf(fp, " %1X", dump[i]);
		}
	}	
	fprintf(fp, "\r\n");
}



/*
 
 */
Opcode *chain_nibble_list(Opcode *opcode, unsigned int nibble) {
	
	Opcode_nibble *nibble_ptr, *ptr;
	
	nibble_ptr = malloc(sizeof(Opcode_nibble));
	nibble_ptr->data = nibble;
	nibble_ptr->next = NULL;
		
	if (opcode->u.opcode_nibble_dn.next == NULL) {
		opcode->u.opcode_nibble_dn.next = nibble_ptr;	
	} else {
	
		ptr = opcode->u.opcode_nibble_dn.next;
		while (ptr->next != NULL) {
			ptr = ptr->next;
		}	
		ptr->next = nibble_ptr;		
	}
	
	ptr = opcode->u.opcode_nibble_dn.next;	
	while (ptr->next != NULL) {
		ptr = ptr->next;
	}
	
	return opcode;
}

/*
 
 */
void print_dump_data() {
	Opcode *ptr;
	unsigned int num;
	Opcode_nibble *nibble;
	Boolean flag = FALSE;
	ptr = Top;
	
	while (ptr != NULL) {		
		switch (ptr->type) {
			case TYPE_NIBBLE_DN:
				flag = TRUE;
				num = 1;

				printf("%02x: ", ptr->u.opcode_nibble_dn.adress);

				nibble = ptr->u.opcode_nibble_dn.next;
				printf("%1x ", ptr->u.opcode_nibble_dn.data);
				while (nibble != NULL) {					
					if ((num % 8) == 0) print_space(1);
					printf("%1x ", nibble->data);
					num++;
					nibble = nibble->next;
				}				
				printf("\n");
				break;
			default:
				break;
		}
		ptr = ptr->next;
	}	
	if (flag == TRUE) printf("\n");
}