
%{

#include "gmc4asm.h"
#include "y.tab.h"

	extern int gmc4_asm_lex_opcode1();
	extern int gmc4_asm_lex_opcode2();
	extern int gmc4_asm_lex_opcode_jump();
	extern int gmc4_asm_lex_opcode_call();
	extern int gmc4_asm_lex_opcode_call_exec();
	extern int gmc4_asm_lex_opcode_align();
	
	int yywrap(void)
	{
		return 1;
	}

%}


DIGIT			[0-9]
ID				[_a-zA-Z][_a-zA-Z0-9]*
BIT4			[0-9a-fA-F]

KA				[kK][aA]
AO				[aA][oO]
CH				[cC][hH]
CY				[cC][yY]
AM				[aA][mM]
MA				[mM][aA]
M_PLUS			([mM]\+)
M_MINUS			([mM]\-)

TIA				[tT][iI][aA]
AIA				[aA][iI][aA]
TIY				[tT][iI][yY]
AIY				[aA][iI][yY]
CIA				[cC][iI][aA]
CIY				[cC][iI][yY]
JUMP			[jJ][uU][mM][pP]

CAL				([cC][aA][lL])|([cC][aA][lL][lL])
	
RSTO			[rR][sS][tT][oO]
SETR			[sS][eE][tT][rR]
RSTR			[rR][sS][tT][rR]
CMPL			[cC][mM][pP][lL]
CHNG			[cC][hH][nN][gG]
SIFT			([sS][iI][fF)][tT])|([sS][hH][iI][fF)][tT])
ENDS			[eE][nN][dD][sS]
ERRS			[eE][rR][rR][sS]
SHTS			[sS][hH][tT][sS]
LONS			[lL][oO][nN][sS]
SUND			[sS][uU][nN][dD]
TIMR			[tT][iI][mM][rR]
DSPR			[dD][sS][pP][rR]
DEM_PLUS		([dD][eE][mM]\+)
DEM_MINUS		([dD][eE][mM]\-)

ALIGN			(\.[aA][lL][iI][gG][nN])
DN				[dD][nN]

LABEL_ID		(\**{ID}\:*)

%%

[ \t]+			/* eat up whitespace */

\;[^\r\n]*		/* eat up comment */

\/\/[^\r\n]*	/* eat up comment */

\#[^\r\n]*		/* eat up comment */

\r				return CR;

\n				return LF;

,				return COMMA;

{BIT4}{BIT4}	{
/*	printf("BIT8: %s\n", yytext);*/
	sscanf(yytext, "%x", &(yylval.value));
	return TOKEN_BIT8;
}


{BIT4}		{
/*	printf("BIT4: %s\n", yytext);*/
	sscanf(yytext, "%x", &(yylval.value));
	return TOKEN_BIT4;
}

{KA}			|
{AO}			|
{CH}			|
{CY}			|
{AM}			|
{MA}			|
{M_PLUS}		|
{M_MINUS}		{
/*	gmc4_asm_lex_opcode1();*/
	return gmc4_asm_lex_opcode1();
}

{TIA}			|
{AIA}			|
{TIY}			|
{AIY}			|
{CIA}			|
{CIY}			{
	return gmc4_asm_lex_opcode2();
}

{JUMP}			{
	return gmc4_asm_lex_opcode_jump();
}

{CAL}			{
	return gmc4_asm_lex_opcode_call();
}

{RSTO}			|
{SETR}			|
{RSTR}			|
{CMPL}			|
{CHNG}			|
{SIFT}			|
{ENDS}			|
{ERRS}			|
{SHTS}			|
{LONS}			|
{SUND}			|
{TIMR}			|
{DSPR}			|
{DEM_PLUS}		|
{DEM_MINUS}		{
	return gmc4_asm_lex_opcode_call_exec();
}

{DN} {
	return TOKEN_OPCODE_NIBBLE_DN;
}

{ALIGN}	{
/*	printf("Align: %s\n", yytext);*/
	return gmc4_asm_lex_opcode_align();
}

{LABEL_ID}	{
/*	printf("Label ID: %s\n", yytext);*/
	char *name;
	
	
	name = name_label_malloc(yytext);
	yylval.name = name;
	
	return TOKEN_LABEL_ID;
}

.           printf( "Unrecognized character: %s\n", yytext );

%%

/*
 
 */
int gmc4_asm_lex_opcode1() {
/*	printf("%s\n", yytext);*/
	
	Opcode *opcode;
	char *name;
	
	name = name_malloc(yytext);

	opcode = opcode_malloc();
	init_opcode(opcode, TYPE_OPCODE1, name);
	
	yylval.opcode = opcode;
	
	return TOKEN_OPCODE1;	
}

/*
 
 */
int gmc4_asm_lex_opcode2() {
/*	printf("%s\n", yytext);*/

	Opcode *opcode;
	char *name;
	
	name = name_malloc(yytext);
	opcode = opcode_malloc();
	init_opcode(opcode, TYPE_OPCODE2, name);
		
	yylval.opcode = opcode;
	
	return TOKEN_OPCODE2;

}

/*
 
 */
int gmc4_asm_lex_opcode_jump() {
/*	printf("%s\n", yytext);*/

	Opcode *opcode;
	
	opcode = opcode_malloc();
	init_opcode(opcode, TYPE_JUMP, NULL);
	
	yylval.opcode = opcode;
	
	return TOKEN_OPCODE_JUMP;
}

/*
 
 */
int gmc4_asm_lex_opcode_call() {
/*	printf("%s\n", yytext);*/
	
	Opcode *opcode;
	
	opcode = opcode_malloc();
	init_opcode(opcode, TYPE_CALL, NULL);
	
	yylval.opcode = opcode;
	
	return TOKEN_OPCODE_CALL;
}

/*
 
 */
int gmc4_asm_lex_opcode_call_exec() {
/*	printf("call %s\n", yytext);*/
	char *name;
	
	
	name = name_malloc(yytext);
	yylval.name = name;
	
	return TOKEN_OPCODE_CALL_EXEC;
}

/*
 
 */
int gmc4_asm_lex_opcode_align() {
	/*	printf("%s\n", yytext);*/
	
	Opcode *opcode;
	
	opcode = opcode_malloc();
	init_opcode(opcode, TYPE_ALIGN, NULL);
	
	yylval.opcode = opcode;
	
	return TOKEN_OPCODE_ALIGN;	
}

