%{
	#include "gmc4asm.h"
	int yylex (void);
	void yyerror (char const *);
%}

%union {
	unsigned int value;
	char *name;
	Opcode *opcode;
}

%token <opcode> TOKEN_OPCODE1
				TOKEN_OPCODE2
%token <opcode> TOKEN_OPCODE_JUMP 
%token <opcode> TOKEN_OPCODE_CALL
%token <opcode> TOKEN_OPCODE_NIBBLE_DN
%token <opcode> TOKEN_OPCODE_ALIGN

%token <name>	TOKEN_OPCODE_CALL_EXEC

%token <name>	TOKEN_LABEL_ID
%token <value>	CR LF COMMA
%token <value>  TOKEN_BIT4
%token <value>  TOKEN_BIT8

%type <value>	input
%type <opcode>	line
%type <opcode>	paragraph
%type <opcode>	opcode
%type <opcode>	nibble_list

%%

input: /* empty */ {}
	| input paragraph
	{
		
		if ($2->type == TYPE_NULL) {
			opcode_free($2);
		} else if (Top == NULL) {
			Top = $2;
		} else {
			add_top($2);
		}
	}
	;


paragraph: line
	{
		$1->line_number = line_number;
		$$ = $1;
	}
	| line CR
	{
		line_number++;
		$1->line_number = line_number;

		$$ = $1;
	}
	| line LF
	{
		line_number++;
		$1->line_number = line_number;

		$$ = $1;
	}
	| line CR LF
	{
		line_number++;
		$1->line_number = line_number;

		$$ = $1;
	}
	;


line:
	{
		Opcode *op;
						
		op = opcode_malloc();
		init_opcode(op, TYPE_NULL, NULL);
			
		$$ = op;
	}
	| TOKEN_BIT8
	{
		yyerror("syntax error");
		YYABORT;
	}
	| TOKEN_BIT4
	{
		yyerror("syntax error");
		YYABORT;
	}
	| TOKEN_OPCODE_ALIGN TOKEN_BIT8
	{
		$1->u.opcode_align.adress = $2;
		$$ = $1;
	}
	| TOKEN_OPCODE_NIBBLE_DN nibble_list
	{
		$$ = $2;
	}
	| opcode
	{
		$$ = $1;
	}
	| TOKEN_LABEL_ID
	{
		Opcode *op;

		op = opcode_malloc();
		init_opcode(op, TYPE_LABEL, $1);

		$$ = op;
	}
	| TOKEN_LABEL_ID opcode
	{		
		$2->label.flag = TRUE;
		$2->label.name = $1;

		$$ = $2;
	}
    ;

opcode:	TOKEN_OPCODE1 { $$ = $1; }
	| TOKEN_OPCODE2 TOKEN_BIT4 
	{

		$1->u.opcode2.code2 = $2;	
		$$ = $1;

		/*	printf("%d\n", $2);*/

	}
	| TOKEN_OPCODE_JUMP TOKEN_LABEL_ID
	{
		$1->u.opcode_jump.name = $2;
		$$ = $1;
	}		
	| TOKEN_OPCODE_CALL TOKEN_OPCODE_CALL_EXEC
	{
		$1->u.opcode_call.name = $2;
		$1->u.opcode_call.code2 = getsym($2);
		$$ = $1;
	}
	;

nibble_list: TOKEN_BIT4
	{
		Opcode *opcode;
		
		opcode = opcode_malloc();
		init_opcode(opcode, TYPE_NIBBLE_DN, NULL);
		opcode->u.opcode_nibble_dn.data = $1;
		
		$$ = opcode;
	}
	| nibble_list COMMA TOKEN_BIT4
	{
		$$ = chain_nibble_list($1, $3);
	}
	; 

/*
exp:      NUM           { $$ = $1;           }
             | exp exp '+'   { $$ = $1 + $2;      }
             | exp exp '-'   { $$ = $1 - $2;      }
             | exp exp '*'   { $$ = $1 * $2;      }
             | exp exp '/'   { $$ = $1 / $2;      }
             | exp exp '^'   { $$ = pow ($1, $2); }
             | exp 'n'       { $$ = -$1;          }
     ;
*/
 
%%

void yyerror (char const *s)
{
	fprintf (stderr, "line %d: %s\n", line_number + 1, s);
}
