#include "gmc4asm.h"

char *Usage[] = {
	"gmc4asm version 1.0.0",
	"Copyright (C) 2009 KOYAMA Yoshiaki.",
	"Usage: gmc4asm [OPTIONS] FILE",
	"OPTIONS",
	"     -n : Output data of normal format.",
	"     -d : Output data of dump format.",
	"     -da: Output all data of dump format.",
	"     -f : Generate simulator file (.fxp).",
	"     -h : Generate simulator file (.hex).",
	NULL
};

void print_usage() {
	int i;
	
	for (i=0; Usage[i] != NULL; i++) {
		printf("%s\n", Usage[i]);
	}	
}

/*
 
 */
int main (int argc, char * argv[]) {

    extern int yyparse(void);
	extern int yylex(void);
	extern FILE *yyin;
	FILE *fp;
	char *file_name, *ptr_name, fxp_name[256], hex_name[256];

	Boolean calc_flag, set_flag, align_flag;
	Boolean dump_flag, dump_all_flag, print_flag, fxp_flag, hex_flag;
	int i;
	size_t len, ll;

	unsigned int dump[STORAGE_MAX];
	unsigned int size;

	/*
	 
	 */
    if (argc == 1) {
		print_usage();
		exit(1);
    }
	
	dump_flag = FALSE;
	dump_all_flag = FALSE;
	print_flag = FALSE;
	fxp_flag = FALSE;
	hex_flag = FALSE;
	
	for (i=1; i<argc; i++) {
		
		if (argv[i][0] == '-') {
			if (strcmp("-d", argv[i]) == 0) {
				dump_flag = TRUE;
			}
			if (strcmp("-da", argv[i]) == 0) {
				dump_all_flag = TRUE;
			}
			if (strcmp("-n", argv[i]) == 0) {
				print_flag = TRUE;
			}
			if (strcmp("-f", argv[i]) == 0) {
				fxp_flag = TRUE;
			}
			if (strcmp("-h", argv[i]) == 0) {
				hex_flag = TRUE;
			}
		} else {	
			file_name = argv[i];
		}
	}
	
	
	fp = fopen(file_name, "r");
    if (fp == NULL) {
		print_usage();
        fprintf(stderr, "\nファイル %s がありません。\n", file_name);
        exit(1);
    }
		
    yyin = fp;

	/*	
    if (yylex()) {
        fprintf(stderr, "Error!\n");
        exit(1);
    }
	 */	

	init_table();
    if (yyparse()) {
        exit(1);
    }

	
	calc_flag = calc_label();
	set_flag = set_label();
	align_flag = set_align();
	
	if (calc_flag == FALSE || set_flag == FALSE || align_flag == FALSE) {
		exit(1);
	}	

	// 表示
	if (((dump_flag == FALSE) && (dump_all_flag == FALSE) && (print_flag == FALSE) && (fxp_flag == FALSE)) || (print_flag == TRUE)) {
		print_top();	
	}
	// ダンプファイル形式の表示

	if (fxp_flag || dump_flag || hex_flag || dump_all_flag) {
		size = calc_dump(dump);
	}

	/*
	
	*/


	if (fxp_flag || hex_flag) {
		len = strlen(file_name);
		ptr_name = file_name;
		
		for (ll=0; ll<len; ll++) {
			
			if ( *ptr_name == '.') {
				fxp_name[ll] = '\0';
				hex_name[ll] = '\0';
				break;		
			}
			
			fxp_name[ll] = *ptr_name;
			hex_name[ll] = *ptr_name;

			ptr_name++;
			
			if (ll == (len-1)) {
				fxp_name[len] = '\0';
				hex_name[len] = '\0';
			}
		}		
	}
	
	
	
	if (fxp_flag) {
		strcat(fxp_name, ".fxp");
	
		fp = fopen(fxp_name, "w");
		
		if (fp == NULL) {
			print_usage();
			fprintf(stderr, "\nファイル %s を開くことができません。\n", fxp_name);
			exit(1);
		} else {		
			print_fxp(fp, dump, size);
			fclose(fp);
		}
	}

	if (hex_flag) {
		strcat(hex_name, ".hex");
		
		fp = fopen(hex_name, "w");
		
		if (fp == NULL) {
			print_usage();
			fprintf(stderr, "\nファイル %s を開くことができません。\n", hex_name);
			exit(1);
		} else {		
			print_hex(fp, dump, size);
			fclose(fp);
		}
	}
	
	if (dump_flag) {
		print_dump(dump, size);
		print_dump_data();
	}
	
	if (dump_all_flag) print_dump(dump, STORAGE_MAX);

	//	print_table();
	
	return 0;
}